/*
 *	program:	pv  -   "particle viewer"
 *
 *      usage:     pv [n]
 *      n = particle #
 *
 *	Areez's n-particle simulator.
 *
 *	Alex Barnett  12/21/01
 *      updated to solid spheres 8/14/04
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <fstream.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>


/* 2^31 - 1 written as a float */
#define MAXINT		2147483647.0
#define PI		3.14159265358979323846
#define max(a,b)	((a) > (b) ? (a) : (b))
#define min(a,b)	((a) < (b) ? (a) : (b))

// number of particles
int n;

// coords
double *x,*y,*z,*vx,*vy,*vz;

// charge
double *ch;

// mass
double *m;

// colour `coords'
float *cx, *cy, *cz;

// energy, potential, interaction energy etc. Other globals
double E, U, V;

// mouse params for L+M adjust
double v,w;

// timestep
double dt;

// iteration number
int its;

// window size, position
int width, height, xpos, ypos;

// old mouse position
int omx, omy;

int helptoggle;
int displaymode;
int axesmode;
int dropmode;

// button info
int left = 0, middle = 0, right = 0;

// eye and angle info
float ex, ey, ez;
float dist;
float th, ph, rh, xof, yof;
// 2 * tangent of half-width of view in Y-direction...
float iris;
float zscale;
float thdot, phdot;


//  ........................ taken from opengl/examples/libaux/shape.c :

#define SPHEREWIRE      0
#define SPHERESOLID     10

/*      structure for each geometric object     */
typedef struct model {
    GLuint list;        /*  display list to render object   */
    struct model *ptr;  /*  pointer to next object      */
    int numParam;       /*  # of parameters             */
    GLdouble *params;   /*  array with parameters       */
} MODEL, *MODELPTR;

static MODELPTR lists[25] = {
    NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL
};

int compareParams (GLdouble *oneArray, GLdouble *twoArray, int size) 
{
    int i;
    int matches = 1;

    for (i = 0; (i < size) && matches; i++) {
        if (*oneArray++ != *twoArray++)
            matches = 0;
    }
    return (matches);
}

GLuint findList (int index, GLdouble *paramArray, int size) 
{
    MODELPTR endList;
    int found = 0;
    
    endList = lists[index];
    while (endList != NULL) {
        if (compareParams (endList->params, paramArray, size))
            return (endList->list);
        endList = endList->ptr;
    }
/*  if not found, return 0 and calling routine should
 *  make a new list     
 */
    return (0);
}

GLuint makeModelPtr (int index, GLdouble *sizeArray, int count)
{
    MODELPTR newModel;

    newModel = (MODELPTR) malloc (sizeof (MODEL));
    newModel->list = glGenLists (1);
    newModel->numParam = count;
    newModel->params = sizeArray;
    newModel->ptr = lists[index];
    lists[index] = newModel;

    return (newModel->list);
}

/*  Render wire frame or solid sphere.  If no display list with
 *  the current model size exists, create a new display list.
 */
void auxWireSphere (GLdouble radius)
{
    GLUquadricObj *quadObj;
    GLdouble *sizeArray;
    GLuint displayList;

    sizeArray = (GLdouble *) malloc (sizeof (GLdouble) * 1);
    *sizeArray = radius;
    displayList = findList (SPHEREWIRE, sizeArray, 1);

    if (displayList == 0) {
        glNewList(makeModelPtr (SPHEREWIRE, sizeArray, 1),
            GL_COMPILE_AND_EXECUTE);
            quadObj = gluNewQuadric ();
            gluQuadricDrawStyle (quadObj, GLU_LINE);
            gluSphere (quadObj, radius, 16, 16);
        glEndList();
    }
    else {
        glCallList(displayList);
        free (sizeArray);
    }
}

void auxSolidSphere (GLdouble radius)
{
    GLUquadricObj *quadObj;
    GLdouble *sizeArray;
    GLuint displayList;

    sizeArray = (GLdouble *) malloc (sizeof (GLdouble) * 1);
    *sizeArray = radius;
    displayList = findList (SPHERESOLID, sizeArray, 1);

    if (displayList == 0) {
        glNewList(makeModelPtr (SPHERESOLID, sizeArray, 1),
            GL_COMPILE_AND_EXECUTE);
            quadObj = gluNewQuadric ();
            gluQuadricDrawStyle (quadObj, GLU_FILL);
            gluQuadricNormals (quadObj, GLU_SMOOTH);
            gluSphere (quadObj, radius, 16, 16);
        glEndList();
    }
    else {
        glCallList(displayList);
        free (sizeArray);
    }
}
// .........................................................................


void light_on(void)
{
    GLfloat light_ambient[] = { 0.0, 0.0, 0.0, 1.0 };
    GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
/*      light_position is NOT default value     */
    GLfloat light_position[] = { 1.0, 0.0, 0.0, 0.0 };
    GLfloat global_ambient[] = { 0.75, 0.75, 0.75, 1.0 };

    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
    
    glFrontFace (GL_CW);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
}

void light_off(void)
{
  glDisable(GL_LIGHTING);
}



// help information
char *helptext[] = {

  " Areez particle viewer",
  "",
  " Left+Middle buttons adjust v,w",
  " v = ...",
  " w = ...",
  "",
  " Left - adj view",
  " Middle - view distance",
  " <space> cycle through display modes",
  " A - cycle through axes modes",
  " D - cycle drops to box walls",
  " Q - quit",
  " Z - restart simulation",
  "",
  "  Press C to continue...",
  NULL };






// --------------------------evolve one timestep----------------------------

#define REG_DIST 0.03

void evolve()
{
  int i,j;
  double *fx, *fy, *fz, fxij, fyij, fzij, dx, dy, dz, r, rrr;


  // update positions
  for (i=0;i<n;++i) {
    x[i] += dt * vx[i];
    y[i] += dt * vy[i];
    z[i] += dt * vz[i];
  }

  fx = new double[n];
  fy = new double[n];
  fz = new double[n];

  // zero forces and interaction energy
  for (i=0;i<n;++i)
    fx[i] = fy[i] = fz[i] = 0.0;
  U = 0.0;

  // sum forces and potential
  for (i=0;i<n;++i)
    for (j=i+1;j<n;++j) {
      dx = x[i]-x[j];
      dy = y[i]-y[j];
      dz = z[i]-z[j];
      r = sqrt(dx*dx + dy*dy + dz*dz);
      // rrr = r*r*r; unregularized coulomb
      // regularized
      rrr = r*(r + REG_DIST)*(r + REG_DIST);
      // force
      fxij = ch[i]*ch[j]*dx/rrr;
      fyij = ch[i]*ch[j]*dy/rrr;
      fzij = ch[i]*ch[j]*dz/rrr;
      // j pulls on i
      fx[i] += fxij;
      fy[i] += fyij;
      fz[i] += fzij;
      // i pulls on j
      fx[j] -= fxij;
      fy[j] -= fyij;
      fz[j] -= fzij;
      // interaction energy, regularized
      U += ch[i]*ch[j]/(r + REG_DIST);
    }

  // add gradient of V to forces
  for (i=0;i<n;++i) {
    // insert form of grad V here
    fx[i] += 0.0;
    fy[i] += ch[i]*w;
    fz[i] += 0.0;
  }

  // update velocities
  for (i=0;i<n;++i) {
    vx[i] += dt * fx[i]/m[i];
    vy[i] += dt * fy[i]/m[i];
    vz[i] += dt * fz[i]/m[i];
  }

  delete fx;
  delete fy;
  delete fz;

  E = V = 0.0;
  // sum kinetic and potential energies
  for (i=0;i<n;++i) {
    E += 0.5*m[i]*(vx[i]*vx[i] + vy[i]*vy[i] + vz[i]*vz[i]);
    // insert form of V here
    V += -ch[i]*y[i]*w;
  }

  ++its;
}


// ------------------------initialize evolution-----------------------------

#define INIT_RANGE 1.0

void initialize()
{
  int i;

  for (i=0;i<n;++i) {
    // pos, mom
    x[i] = INIT_RANGE*(random()/MAXINT - 0.5);
    y[i] = INIT_RANGE*(random()/MAXINT - 0.5);
    z[i] = INIT_RANGE*(random()/MAXINT - 0.5);
    vx[i] = 0.5*(random()/MAXINT - 0.5);
    vy[i] = 0.5*(random()/MAXINT - 0.5);
    vz[i] = 0.5*(random()/MAXINT - 0.5);

    // charge
    ch[i] = (random() & 0x01) ? 1.0 : -1.0;

    // mass
    m[i] = 1.0;

    // colours
    cx[i] = (ch[i]>0.0) ? 1.0 : 0.0;
    cy[i] = (ch[i]>0.0) ? 0.0 : 1.0;
    cz[i] = 0.3;
   
    /* colour by particle number
      cx[i] = i/(float)n;
      cy[i] = 1.0 - i/(float)n;
      cz[i] = 0.3;
    */
  }

  dt = v;
  its = 0;

}


// ------------------------allocate-----------------------------
void allocate()
{
  x = new double[n];
  y = new double[n];
  z = new double[n];
  vx = new double[n];
  vy = new double[n];
  vz = new double[n];
  ch = new double[n];
  m = new double[n];
  cx = new float[n];
  cy = new float[n];
  cz = new float[n];
}


// ------------------------deallocate-----------------------------
void deallocate()
{
  delete x;
  delete y;
  delete z;
  delete vx;
  delete vy;
  delete vz;
  delete ch;
  delete m;
  delete cx;
  delete cy;
  delete cz;
}


// ------------------------initialize view--------------------------------
void setviewcoords()
{
  // modes
  displaymode = 1;
  axesmode = 3;
  helptoggle = 0;
  dropmode = 0;

  // eye
  dist = 10.0;
  th = 0.4;
  ph = 0.7;
  rh = 0;
  xof = 0;
  yof = 0;
  thdot = 0.0;
  phdot = 0.0;
  iris = 0.6;
  ex = xof + dist*sin(th)*cos(ph);
  ey = yof + -dist*cos(th)*cos(ph);
  ez = -dist*sin(ph);
}



#define BOX_SIZE_X 3.0
#define BOX_SIZE_Y 2.0
#define BOX_SIZE_Z 1.0

// -----------------------------------------------------------------------

#define N_AXES_MODES 4

void axes()
  // plot 3d axes
{
  //glEnable(GL_LINE_SMOOTH);
  glLineWidth(1.0);


  // box:
  if (axesmode>0) {
    glColor3f(0.3,0.3,1.0); // blue
    glPushMatrix();
    glScalef(BOX_SIZE_X, BOX_SIZE_Y, BOX_SIZE_Z);
    glBegin(GL_LINE_LOOP);
    glVertex3f(1.0,1.0,1.0);
    glVertex3f(1.0,-1.0,1.0);
    glVertex3f(-1.0,-1.0,1.0);
    glVertex3f(-1.0,1.0,1.0);
    glEnd();
    glBegin(GL_LINE_LOOP);
    glVertex3f(1.0,1.0,-1.0);
    glVertex3f(1.0,-1.0,-1.0);
    glVertex3f(-1.0,-1.0,-1.0);
    glVertex3f(-1.0,1.0,-1.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(1.0,1.0,1.0);
    glVertex3f(1.0,1.0,-1.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(1.0,-1.0,1.0);
    glVertex3f(1.0,-1.0,-1.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(-1.0,-1.0,1.0);
    glVertex3f(-1.0,-1.0,-1.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(-1.0,1.0,1.0);
    glVertex3f(-1.0,1.0,-1.0);
    glEnd();
    glPopMatrix();
  }

  if (axesmode>1) {
    // axes
    glColor3f(0.3,0.8,1.0); // cyan
    glBegin(GL_LINE_STRIP);
    glVertex3f(1.0,0.0,0.0);
    glVertex3f(0.0,0.0,0.0);
    glVertex3f(0.0,1.0,0.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(0.0,0.0,0.0);
    glVertex3f(0.0,0.0,1.0);
    glEnd();
  }

  if (axesmode>2) {
    // xyz
    glColor3f(0.3,1.0,1.0); // cyan
    glRasterPos3f(1.1,0.0,0.0);
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'x');
    glRasterPos3f(0.0,1.1,0.0);
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'y');
    glRasterPos3f(0.0,0.0,1.1);
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'z');
  }
}


// ----------------------------------------------------------
void texto(GLshort x, GLshort y, char *text)
{
  char *p;

  glRasterPos2s(x,y);
  for (p = text; *p; p++)
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *p);
}

void fancytexto(GLshort x, GLshort y, char *text)
{
  char *p;

  glRasterPos2s(x,y);
  for (p = text; *p; p++)
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *p);
}


// -----------------------------------------------------------------------
void decorate()
/* 2D text plotting + pixel image */
 /* glMatrixMode(GL_MODELVIEW); */
{
  char str[100];
  int i;

  // 2d projection
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0, width-0.5, 0.0, height-0.5);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if (helptoggle) {
    // display help
    glColor3f(1.0,0.8,0.2);
    for (i = 0; helptext[i]; i++)
      fancytexto(16, height - 50 - i * 24, helptext[i]);
  }

  // output params on screen

  glColor3f(1.0,0,1.0);
  sprintf(str," its=%d  E=%.6g  V=%.6g  U=%.6g  tot=%.6g",its,E,V,U,E+V+U);
  texto(width-500, height-20, str);

  glColor3f(0,1.0,0);
  sprintf(str," disp:%d  axes:%d  drop:%d  v=%g  w=%g",\
	  displaymode,axesmode,dropmode,v,w);
  texto(0, 8, str);
}



// -----------------------------------------DISPLAY----------------------

#define N_DISP_MODES 5
#define N_DROP_MODES 2

void display(void)
{
  int i;
  char text[100],*p;

  // update model spin
  th += thdot;
  th = (th > 0.0) ? ((th < 2*PI) ? th : th-2*PI) : th+2*PI;
  ph += phdot;
  ph = (ph > 0.0) ? ((ph < 2*PI) ? ph : ph-2*PI) : ph+2*PI;

  // clear window
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // 3d projection
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(atan(iris/2)*360.0/PI, width/(float)height, 0.05, 20.0);

  // eye location and angle
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(xof,yof,-dist);
  glRotatef(-rh*180.0/PI, 0.0,1.0,0.0);
  glRotatef(-ph*180.0/PI, 1.0,0.0,0.0);
  glRotatef(-th*180.0/PI, 0.0,0.0,1.0);

  //glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  axes();
  //glDisable(GL_BLEND);


  if (displaymode<2) {
    // ........................... 0,1: particles as points in 3d .............
    glEnable(GL_POINT_SMOOTH);
    glPointSize((displaymode==0)?1.0:2.0); // two sizes
    glBegin(GL_POINTS);
    for (i=0; i<n; ++i) {
      glColor3f(cx[i],cy[i],cz[i]);
      glVertex3f((float)x[i],(float)y[i],(float)z[i]);
    }
    glEnd();
  } else if(displaymode==2) {
    // ........................... 2: particles as symbols in 3d .............
    glDisable(GL_POINT_SMOOTH);
    for (i=0; i<n; ++i) {
      // colour = charge
      // glColor3f((ch[i]>0.0)?1.0:0.0,(ch[i]>0.0)?0.0:1.0,0.0);
      // regular colour
      glColor3f(cx[i],cy[i],cz[i]);
      glRasterPos3f((float)x[i],(float)y[i],(float)z[i]);
      sprintf(text,"%d",i);
      for (p = text; *p; p++)
	glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *p);
    }
  } else if (displaymode==3 || displaymode==4) {
    //...................... 3,4: wire, solid spheres ......................
    //light_on();
    for (i=0; i<n; ++i) {    
      glPushMatrix();
      glTranslatef((float)x[i],(float)y[i],(float)z[i]);
      if (displaymode==3)
	auxWireSphere(0.1);
      else
	auxSolidSphere(0.1);
      glPopMatrix();
    }
    //light_off();
    //glDisable(GL_LINE_SMOOTH);
  }

  if (dropmode>0) {
    // draw lines dropping to box base
    glDisable(GL_LINE_SMOOTH);
    glLineWidth(1.0);
    glBegin(GL_LINES);
    for (i=0; i<n; ++i) {
      glColor3f(cx[i],cy[i],cz[i]);
      glVertex3f((float)x[i],(float)y[i],(float)z[i]);
      glVertex3f((float)x[i],(float)y[i],-(float)BOX_SIZE_Z);
    }
    glEnd();
  }

  decorate();

  glutSwapBuffers();

  // do some number of iterations
  for (i=1;i<=3;++i)
    evolve();

  // if not frozen, set up interrupt to call display again
  if (!right)
    glutPostRedisplay();
}



// ------------------------------- MENU options ------------------------------
void menu(int value)
{
  switch (value) {
  case 1:
    setviewcoords();
    break;
  case 9:
    helptoggle = 1 - helptoggle;
    break;
  case 10:
    exit(0);
    break;
  }
}


// -------------------------------------------------------------
void menustat(int status)
{
  if (status == GLUT_MENU_IN_USE) {
    right = 1;
  } else if (status == GLUT_MENU_NOT_IN_USE) {
    right = 0;
    // call display cycle into action again...
    glutPostRedisplay();
  }
}


// ----------------------------- KEYPRESS options ---------------------------
void keyfunc( unsigned char key, int x, int y)
{
  int i,j;
  switch (key) {
  case 'C':
  case 'c':
    helptoggle = 0;
    break;
  case 'I':
  case 'i':
    printf("some output\n");
    break;
  case ' ':
    ++displaymode;
    if (displaymode==N_DISP_MODES)
      displaymode = 0;
    if (displaymode==4)
      light_on();
    else
      light_off();
    break;
  case 'A':
  case 'a':
    ++axesmode;
    if (axesmode==N_AXES_MODES)
      axesmode = 0;
    break;
  case 'D':
  case 'd':
    ++dropmode;
    if (dropmode==N_DROP_MODES)
      dropmode = 0;
    break;
  case 'Z':
  case 'z':
    initialize();
    break;
  case 'q':
    exit(0);
    break;
    
  }
}


// ---------------------------------motion---------------------------------
void motion(int mx,int my)
  // deals with mouse motion mx my are relative to window origin
{
  int d;
  // change origin from lower left (OpenGL, IrisGL) to upper left (GLUT)
  my = height - my;

  if (left) {
    if (middle) {
      // left and middle together...
      /*   xof -= (mx-omx)*dist/100.0;
       *  yof -= (my-omy)*dist/100.0; */
      // make u,v change
      v *= exp((mx-omx)*3/(float)width);
      w *= exp((my-omy)*3/(float)width);
    } else {
      // left alone....
      d = (mx>omx) ? 1 : -1;
      if (mx==omx) d=0;
      thdot = -(mx-omx-d)*2/(float)width;
      d = (my>omy) ? 1 : -1;
      if (my==omy) d=0;
      phdot = (my-omy-d)*2/(float)width;
    }
  } else if (middle) {
    // middle alone....
    dist += (mx-omx)*3/(float)width;
    // dist = (dist > -20.0) ? ((dist < 100.0) ? dist : 100.0) : -20.0;
    iris += (my-omy)/400.0;
    iris = (iris > 0.05)?((iris < 5.0) ? iris : 5.0) : 0.05;
    // zscale *= exp((my-omy)*3/(float)width);
  }

  omx = mx;
  omy = my;
}


// ----------------------------------------------------------------------
void passivemotion( int x, int y)
  // compute new cursor position whenever mouse moves
{
  y = height - y;
}


// ----------------------------------------------------------------------
void mouse(int button, int state, int x, int y)
{
  y = height - y;
  if (button == GLUT_LEFT_BUTTON) {
    if (state == GLUT_DOWN) {
      left = 1;
	omx = x;
	omy = y;
    } else if (state == GLUT_UP) {
      left = 0;
      if (middle)
	// do something when params u,v change
	dt = v;
   }
  }
  if (button == GLUT_MIDDLE_BUTTON) {
    if (state == GLUT_DOWN) {
      middle = 1;
	omx = x;
	omy = y;
    } else if (state == GLUT_UP) {
      middle = 0;
      if (left)
	// do something when params u,v change
	dt = v;
    }
  }
  if (left || middle)
    glutSetCursor(GLUT_CURSOR_INFO);
  else
    glutSetCursor(GLUT_CURSOR_CROSSHAIR);
}


// ----------------------------------------------------------------------
void reshape(int w, int h)
{
  glViewport(0, 0, w, h);
  width = w;
  height = h;
}



// ================================================== MAIN ===============
int main(int argc, char **argv)
{
  int i;

  // read command line
  n = 10;
  if (argc>1) {
	sscanf(argv[1],"%d",&n);
	--argc;
	++argv;
  }


  // initialize mouse-adjustable params
  v = 0.001;  // timestep
  w = 0.3;    // E field in y direc

  allocate();

  initialize();

  setviewcoords();

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(500, 500);
  glutCreateWindow("pv");

  width  = glutGet(GLUT_WINDOW_WIDTH);
  height = glutGet(GLUT_WINDOW_HEIGHT);
  xpos   = glutGet(GLUT_WINDOW_X);
  ypos   = glutGet(GLUT_WINDOW_Y);

  // from examples/teaambient.c:
    glEnable(GL_AUTO_NORMAL);
    glEnable(GL_NORMALIZE);
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);

  glutPassiveMotionFunc(passivemotion);
  glutMotionFunc(motion);
  glutMouseFunc(mouse);
  glutReshapeFunc(reshape);
  glutDisplayFunc(display);
  glutKeyboardFunc(keyfunc);
  glutMenuStateFunc(menustat);

  glutCreateMenu(menu);
  glutAddMenuEntry("reset view", 1);
  glutAddMenuEntry("help", 9);
  glutAddMenuEntry("quit", 10);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  // pass control to GLUT
  glutMainLoop();

  return 0;             // ANSI C requires main to return int
}

// end
